from ucimlrepo import fetch_ucirepo


def fetch_cars_dataset():
    """
    Fetch the cars dataset from the UCI repository.
    """
    uci = fetch_ucirepo(id=9)
    X = uci.data.features
    y = uci.data.targets["mpg"]

    null_indices = X[X.isnull().any(axis=1)].index.union(y[y.isnull()].index)
    X = X.drop(null_indices).reset_index(drop=True)
    y = y.drop(null_indices).reset_index(drop=True)

    return X, y


def fetch_wine_dataset():
    """
    Fetch the wine dataset from the UCI repository.
    """
    uci = fetch_ucirepo(id=109)
    X = uci.data.features
    y = uci.data.targets["class"]

    X.columns = [x.lower() for x in X.columns]

    return X, y



"""Functions for manipulating the symbolic programs."""

from typing import List

import numpy as np
import pandas as pd

from .symbolic_functions import SymbolicFunction


def random_prog(depth: int, X: pd.DataFrame, operations: List[SymbolicFunction]):
    """
    Generate a random program for symbolic regression.

    Parameters
    ----------
    depth : int
        The depth of the program.

    X : pd.DataFrame
        The input data.

    operations : list
        The list of operations to use.
    """
    if np.random.randint(0, 10) >= depth * 2:
        op = operations[np.random.randint(0, len(operations) - 1)]()
        return {
            "func": op,
            "children": [
                random_prog(depth + 1, X, operations) for _ in range(op.arg_count)
            ],
            "format_str": op.format_str,
        }

    return {"feature_name": X.columns[np.random.randint(0, X.shape[1] - 1)]}


def node_count(node: dict) -> int:
    """
    Count the number of nodes in the program.

    Parameters
    ----------
    node : dict
        The program to count.

    Returns
    -------
    int
        The number of nodes in the program.
    """
    if "children" not in node:
        return 1
    return sum((node_count(c) for c in node["children"]))


def select_random_node(selected: dict, parent: dict, depth: int) -> dict:
    """
    Select a random node from the program.

    Parameters
    ----------

    selected : dict
        The current node.

    parent : dict
        The parent node.

    depth : int
        The depth of the program.

    Returns
    -------
    dict
        The selected node.
    """
    if "children" not in selected:
        return parent

    if np.random.randint(0, 10) < 2 * depth:
        return selected

    child_count = len(selected["children"])
    child_idx = 0 if child_count <= 1 else np.random.randint(0, child_count - 1)

    return select_random_node(
        selected["children"][child_idx],
        selected,
        depth + 1,
    )


def render_prog(node: dict) -> str:
    """
    Render a program to a string.

    Parameters
    ----------
    node : dict
        The program to render.

    Returns
    -------
    str
        The rendered program.
    """
    if "children" not in node:
        return node["feature_name"]
    return node["format_str"].format(*[render_prog(c) for c in node["children"]])



import featuristic as ft
import pandas as pd


def test_preprocess():
    X = pd.DataFrame({"a": [1, 1, 1], "b": [4, 5, 6]})
    X_new = ft.synthesis.preprocess.drop_low_variance_features(X, threshold=0.1)
    assert X_new.columns.tolist() == ["b"]


def test_preprocess_data():
    X = pd.DataFrame({"a": [1, 1, 1], "b": [4, 5, 6]})
    y = pd.Series([1, 2, 3])
    X_new, y_new = ft.synthesis.preprocess.preprocess_data(X, y)
    assert X_new.columns.tolist() == ["b"]
    assert y_new.tolist() == [1, 2, 3]



""" The module to represent the population of programs in the genetic programming algorithm. """

import copy
import sys
from typing import Callable, List, Tuple

import numpy as np
import pandas as pd
from joblib import Parallel, cpu_count, delayed


class BasePopulation:
    """
    A class to represent the population of programs in the genetic programming algorithm.

    Parameters
    ----------
    population_size : int
        The size of the population.
    feature_count : int
        The number of features in the dataset.
    crossover_proba : float, optional
        The probability of crossover, by default 0.9.
    mutation_proba : float, optional
        The probability of mutation, by default 0.1.
    """

    def __init__(
        self,
        population_size: int,
        feature_count: int,
        crossover_proba: float = 0.9,
        mutation_proba: float = 0.1,
    ):
        self.population_size = population_size
        self.feature_count = feature_count
        self.crossover_proba = crossover_proba
        self.mutation_proba = mutation_proba

        self.population = None
        self._initialize_population()

    def _initialize_population(self):
        """
        Initialize the population.

        Returns
        -------
        np.ndarray:
            The initial population.
        """
        self.population = np.random.choice(
            [0, 1], size=(self.population_size, self.feature_count)
        )

    def evaluate(
        self, cost_func: Callable, X: pd.DataFrame, y: pd.Series
    ) -> List[pd.Series]:
        """
        Evaluate the population against the dataframe of features.

        Args
        ----
        cost_func : Callable
            The cost function to evaluate the individual's fitness.
        X : pd.DataFrame
            The dataframe with the features.
        y : pd.Series
            The true values.

        Returns
        -------
        List[pd.Series]
            The predicted values.
        """
        raise NotImplementedError

    def _evaluate(
        self, cost_func: Callable, X: pd.DataFrame, y: pd.Series, genome: np.ndarray
    ) -> float:
        """
        Evaluate the populations's fitness using the cost function.

        Parameters
        ----------
        cost_func : Callable
            The cost function to evaluate the individual's fitness.
        X : pd.DataFrame
            The dataframe with the features.
        y : pd.Series
            The true values.
        genome : np.ndarray
            The genome of an individual.

        Returns
        -------
        float
            The fitness of the individual.
        """
        if genome.sum() == 0:
            current_cost = sys.maxsize
        else:
            current_cost = cost_func(X[X.columns[genome == 1]], y)
        return current_cost

    def _selection(self, scores: List, k: int = 3) -> np.ndarray:
        """
        Select an individual from the population using tournament selection.

        Parameters
        ----------
        scores : List
            The fitness scores of the population.
        k : int, optional
            The number of individuals to select for the tournament, by default 3.

        Returns
        -------
        np.ndarray
            The selected individual.
        """
        selection_ix = np.random.randint(len(self.population))

        for ix in np.random.randint(0, len(self.population), k - 1):
            if scores[ix] < scores[selection_ix]:
                selection_ix = ix
        return self.population[selection_ix]

    def _mutate(self, genome) -> np.ndarray:
        """
        Mutate the individual's genome.

        Parameters
        ----------
        genome : np.ndarray
            The genome of an individual.

        Returns
        -------
        np.ndarray
            The mutated genome.
        """
        proba = np.random.uniform(size=len(genome))
        mask = proba < self.mutation_proba
        genome[mask == 1] = 1 - genome[mask == 1]
        return genome

    def _crossover(
        self, parent1: np.ndarray, parent2: np.ndarray, crossover_proba: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Crossover two individuals to create two new children.

        Parameters
        ----------
        parent1 : np.ndarray
            The first parent.
        parent2 : np.ndarray
            The second parent.
        crossover_proba : float
            The probability of crossover.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The two new children.
        """
        c1, c2 = copy.deepcopy(parent1), copy.deepcopy(parent2)

        if np.random.rand() < crossover_proba:
            pt = np.random.randint(1, len(parent1) - 2)
            c1 = np.concatenate([parent1[:pt], parent2[pt:]])
            c2 = np.concatenate([parent2[:pt], parent1[pt:]])
        return c1, c2

    def evolve(self, fitness: List[float]):
        """
        Evolve the population based on the fitness scores.

        Parameters
        ----------
        fitness : List[float]
            The fitness scores of the population.
        """
        selected = [self._selection(fitness) for _ in range(self.population_size)]

        children = []
        for i in range(0, len(self.population) - 1, 2):
            p1, p2 = selected[i], selected[i + 1]
            for c in self._crossover(p1, p2, self.crossover_proba):
                c = self._mutate(c)
                children.append(c)

        self.population = children


class SerialPopulation(BasePopulation):
    """
    A class to represent the population of programs in the genetic programming algorithm where
    the programs are evaluated serially.
    """

    def __init__(
        self,
        population_size: int,
        feature_count: int,
        crossover_proba: float = 0.9,
        mutation_proba: float = 0.1,
    ):
        """
        Initialize the population class.

        Args
        ----
        population_size : int
            The size of the population.
        feature_count : int
            The number of features in the dataset.
        crossover_proba : float, optional
            The probability of crossover, by default 0.9.
        mutation_proba : float, optional
            The probability of mutation, by default 0.1.
        """
        super().__init__(
            population_size,
            feature_count,
            crossover_proba,
            mutation_proba,
        )

    def evaluate(
        self, cost_func: Callable, X: pd.DataFrame, y: pd.Series
    ) -> List[float]:
        """
        Evaluate the population against the dataframe of features.

        Args
        ----
        cost_func : Callable
            The cost function to evaluate the individual's fitness.
        X : pd.DataFrame
            The dataframe with the features.
        y : pd.Series
            The true values.

        Returns
        -------
        List[float]
            The predicted values.
        """
        return [self._evaluate(cost_func, X, y, genome) for genome in self.population]


class ParallelPopulation(BasePopulation):
    """
    A class to represent the population of programs in the genetic programming algorithm where
    the programs are evaluated in parallel.
    """

    def __init__(
        self,
        population_size: int,
        feature_count: int,
        crossover_proba: float = 0.9,
        mutation_proba: float = 0.1,
        n_jobs: int = -1,
    ):
        """
        Initialize the population class.

        Args
        ----
        population_size : int
            The size of the population.
        feature_count : int
            The number of features in the dataset.
        crossover_proba : float, optional
            The probability of crossover, by default 0.9.
        mutation_proba : float, optional
            The probability of mutation, by default 0.1.
        n_jobs : int, optional
            The number of parallel jobs to run, by default -1 (use all available cores).
        """
        super().__init__(
            population_size,
            feature_count,
            crossover_proba,
            mutation_proba,
        )

        n_jobs = cpu_count() if n_jobs == -1 else n_jobs

    def evaluate(
        self, cost_func: Callable, X: pd.DataFrame, y: pd.Series
    ) -> List[float]:
        """
        Evaluate the population against the dataframe of features.

        Args
        ----
        cost_func : Callable
            The cost function to evaluate the individual's fitness.
        X : pd.DataFrame
            The dataframe with the features.
        y : pd.Series
            The true values.

        Returns
        -------
        List[float]
            The predicted values.
        """
        return Parallel(n_jobs=cpu_count())(
            delayed(self._evaluate)(cost_func, X, y, genome)
            for genome in self.population
        )



import featuristic as ft
import pandas as pd
import numpy as np


def test_random_prog():
    np.random.seed(0)
    X = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    operations = [
        ft.synthesis.symbolic_functions.SymbolicAdd,
        ft.synthesis.symbolic_functions.SymbolicSubtract,
    ]
    prog = ft.synthesis.program.random_prog(2, X, operations)

    assert isinstance(prog["func"], ft.synthesis.symbolic_functions.SymbolicAdd)
    assert "children" in prog
    assert "format_str" in prog
    assert "feature_name" in prog["children"][0]
    assert prog["children"][0]["feature_name"] in X.columns


def test_node_count():
    np.random.seed(0)
    X = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    node = {
        "func": ft.synthesis.symbolic_functions.SymbolicAdd(),
        "children": [
            {"feature_name": "a"},
            {"feature_name": "b"},
        ],
        "format_str": "({} + {})",
    }
    assert ft.synthesis.program.node_count(node) == 2


def test_select_random_node():
    np.random.seed(0)
    node = {
        "func": ft.synthesis.symbolic_functions.SymbolicAdd(),
        "children": [
            {"feature_name": "a"},
            {"feature_name": "b"},
        ],
        "format_str": "({} + {})",
    }
    selected = ft.synthesis.program.select_random_node(node, None, 2)

    assert type(selected) == dict
    assert type(selected["children"][0]) == dict
    assert "func" in selected


def test_render_prog():
    node = {
        "func": ft.synthesis.symbolic_functions.SymbolicAdd(),
        "children": [
            {"feature_name": "a"},
            {"feature_name": "b"},
        ],
        "format_str": "({} + {})",
    }

    assert ft.synthesis.program.render_prog(node) == "(a + b)"



import featuristic as ft
import pytest
import pandas as pd


def test_gfs():
    num_features = 5
    gfs = ft.GeneticFeatureSynthesis(num_features=num_features, pbar=False)

    with pytest.raises(Exception):
        gfs.plot_history()

    assert gfs.len_hall_of_fame == num_features * 5
    assert gfs.fit_called == False

    with pytest.raises(Exception):
        gfs.fit(X=None, y=None)

    X = pd.DataFrame({"a": [1, 2, 4], "b": [4, 5, 6]})
    y = pd.Series([1, 2, 3])

    gfs.fit(X, y)
    new_X = gfs.transform(X)
    new_cols = [x for x in new_X.columns if x.startswith("feature_")]
    assert len(new_cols) == num_features
    assert gfs.fit_called == True

    gfs = ft.GeneticFeatureSynthesis(num_features=num_features, pbar=False)
    new_X = gfs.fit_transform(X, y)
    new_cols = [x for x in new_X.columns if x.startswith("feature_")]
    assert len(new_cols) == num_features



""" Featuristic: A feature engineering library for machine learning. """

from .datasets import fetch_cars_dataset, fetch_wine_dataset
from .synthesis.genetic_feature_synthesis import GeneticFeatureSynthesis
from .synthesis.symbolic_functions import SymbolicFunction, list_operations
from .selection.genetic_feature_selection import GeneticFeatureSelector
from .version import __version__



""" This module contains the classes and functions for feature synthesis. """

from .genetic_feature_synthesis import GeneticFeatureSynthesis
from .fitness import fitness_pearson
from .mrmr import MaxRelevanceMinRedundancy
from .population import BasePopulation, ParallelPopulation, SerialPopulation
from .program import node_count, random_prog, render_prog, select_random_node
from .symbolic_functions import SymbolicFunction, list_operations



"""Class for selecting most relevant features using the mrmr algorithm."""

from typing import List

import pandas as pd
from sklearn.feature_selection import f_classif, f_regression
from tqdm import tqdm

# set the floor value for the correlation matrix
FLOOR: float = 0.00001


class MaxRelevanceMinRedundancy:
    """
    Class for selecting most relevant features using the mrmr algorithm.
    """

    def __init__(self, k: int = 6, problem_type: str = "regression", pbar=True):
        """
        Initialize the MaxRelevanceMinRedundancy class.

        Parameters
        ----------
        K : int (default=6)
            The number of features to select.

        problem_type : str (default='regression')
            The type of problem. Either 'regression' or 'classification'.

        pbar : bool (default=True)
            Whether to display a progress bar or not.
        """
        self.k = k
        self.selected_features = None
        self.pbar = pbar

        if problem_type not in ["regression", "classification"]:
            raise ValueError(
                "Invalid type. Must be either 'regression' or 'classification'."
            )
        if problem_type == "regression":
            self.metric = f_regression
        else:
            self.metric = f_classif

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """
        Fit the mRMR algorithm to the data.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        Returns
        -------
        None
        """
        self.selected_features = self._mrmr(X, y)

    def transform(self, X: pd.DataFrame, y: pd.Series = None):
        """
        Transform the data using the selected features.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable. Not used in this function.

        Returns
        -------
        pd.DataFrame
            The dataframe with the selected features.
        """
        return X[self.selected_features]

    def fit_transform(self, X: pd.DataFrame, y: pd.Series):
        """
        Fit the mRMR algorithm to the data and transform the data using the selected features.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        Returns
        -------
        pd.DataFrame
            The dataframe with the selected features.
        """
        self.fit(X, y)
        return self.transform(X)

    def _mrmr(self, X: pd.DataFrame, y: pd.Series) -> List[str]:
        """
        Select the top n_features features using the mRMR algorithm.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        Returns
        -------
        list
            The list of selected features.
        """
        # set the maximum number of features to select
        k: int = min(self.k, X.shape[1])

        X = X.loc[:, X.nunique() > 1].dropna(axis=1)

        # calculate the f-statistic and the correlation matrix
        f_stat = pd.Series(self.metric(X, y)[0], index=X.columns)
        corr = pd.DataFrame(FLOOR, index=X.columns, columns=X.columns)

        # initialize list of selected features and list of excluded features
        selected = []
        not_selected = X.columns.to_list()

        if self.pbar:
            pbar = tqdm(total=k, desc="Pruning feature space...")

        # select the top K features
        for i in range(k):
            if i > 0:
                last_selected = selected[-1]
                corr.loc[not_selected, last_selected] = (
                    X[not_selected].corrwith(X[last_selected]).abs().clip(FLOOR)
                )

            score = f_stat.loc[not_selected] / corr.loc[not_selected, selected].mean(
                axis=1
            ).fillna(FLOOR)

            best = score.index[score.argmax()]
            selected.append(best)
            not_selected.remove(best)

            if self.pbar:
                pbar.update(1)

        return selected



import featuristic as ft
import pandas as pd


def test_fetch_cars():
    X, y = ft.fetch_cars_dataset()

    assert X.shape == (392, 7)
    assert y.shape == (392,)
    assert type(X) == pd.DataFrame
    assert type(y) == pd.Series


def test_wine_cars():
    X, y = ft.fetch_wine_dataset()

    assert X.shape == (178, 13)
    assert y.shape == (178,)
    assert type(X) == pd.DataFrame
    assert type(y) == pd.Series



import featuristic as ft
import pytest
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_val_score
import numpy as np


def objective_function(X, y):
    model = LinearRegression()
    scores = cross_val_score(model, X, y, cv=3, scoring="neg_mean_absolute_error")
    return scores.mean() * -1


def test_selection():
    np.random.seed(8888)
    gfs = ft.GeneticFeatureSelector(objective_function)

    with pytest.raises(Exception):
        gfs.fit(X=None, y=None)

    X = pd.DataFrame(
        {"a": [1, 2, 3], "b": [4, 5, 6], "c": [10, 10, 10], "d": [1, 1, 1]}
    )
    y = pd.Series([1, 2, 3])

    gfs.fit(X, y)
    new_X = gfs.transform(X)
    assert new_X.columns.tolist() == ["a", "b"]
    assert new_X.shape[0] == 3
    assert gfs.is_fitted_ == True



import featuristic as ft
import numpy as np
import pandas as pd
import sys
from pytest import approx


def test_pearson_fitness():
    a = pd.Series([1, 2, 3, 4, 5])
    b = pd.Series([1, 2, 3, 4, np.nan])
    c = ft.synthesis.fitness.fitness_pearson({}, 1, a, b)
    assert c == sys.maxsize

    a = pd.Series([1, 2, 3, 4, 5])
    b = pd.Series([1, 2, 3, 4, np.inf])
    c = ft.synthesis.fitness.fitness_pearson({}, 1, a, b)
    assert c == sys.maxsize

    a = pd.Series([1, 1, 1])
    b = pd.Series([1, 1, 1])
    c = ft.synthesis.fitness.fitness_pearson({}, 1, a, b)
    assert c == sys.maxsize

    a = pd.Series([1, 2, 3, 4, 5])
    b = pd.Series([1, 2, 3, 4, 5])
    c = ft.synthesis.fitness.fitness_pearson({}, 0.001, a, b)
    assert c == approx(1.0)



"""Fitness functions for measuring how well the features are performing"""

import warnings

import numpy as np
import pandas as pd
import scipy
import sys
from scipy.stats import pearsonr

from .program import node_count


def fitness_pearson(
    program: dict, parsimony: float, y_true: pd.Series, y_pred: pd.Series
):
    """
    Compute the fitness of a program based on the pearson correlation and the parsimony coefficient

    Args:

    program: dict
        The program to evaluate

    parsimony: float
        The parsimony coefficient

    y_true: pd.Series
        The true values

    y_pred: pd.Series
        The predicted values
    """

    with warnings.catch_warnings(record=True) as _:
        warnings.simplefilter("ignore", category=scipy.stats.NearConstantInputWarning)
        if y_pred.isna().any():
            return sys.maxsize

        if np.isinf(y_pred).any():
            return sys.maxsize

        if np.ptp(y_true) == 0 or np.ptp(y_pred) == 0:
            return sys.maxsize

        loss = abs(pearsonr(y_true, y_pred).statistic)
        penalty = node_count(program) ** parsimony
        loss /= penalty
        return -loss



""" Preprocessing functions for data synthesis. """

from typing import Tuple

import pandas as pd


def preprocess_data(X: pd.DataFrame, y: pd.Series) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Preprocess the input data.

    Parameters
    ----------
    X : pd.DataFrame
        The input data.
    y : pd.Series, default=None
        The target data.
    drop_na : bool, default=True
        Whether to drop rows with missing values.
    drop_low_variance : bool, default=True
        Whether to drop features with low variance.

    Returns
    -------
    pd.DataFrame
        The preprocessed input data.
    pd.Series or None
        The preprocessed target data.
    """
    # drop duplciated columns
    X = X.loc[:, ~X.columns.duplicated()]

    # drop constant columns
    X = X.loc[:, X.nunique() > 1]

    return X, y


def drop_low_variance_features(X, threshold=0.01):
    """
    Drop features with variance below a certain threshold.

    Parameters
    ----------
    X : pd.DataFrame
        The input data.
    threshold : float, default=0.1
        The threshold below which features are dropped.

    Returns
    -------
    pd.DataFrame
        The input data with low-variance features dropped.
    """
    return X.loc[:, X.var() > threshold]



"""Functions to use in the symbolic regression"""

from typing import Callable, List

import numpy as np


def safe_div(a, b) -> np.ndarray:
    """
    Perform safe division by avoiding division by zero.

    Parameters
    ----------
    a : float
        The numerator.

    b : float
        The denominator.

    Returns
    -------
    np.ndarray:
        The result of the division.
    """
    return np.select([b != 0], [a / b], default=a)


def negate(a) -> np.ndarray:
    """
    Negate the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The negated input.
    """
    return np.multiply(a, -1)


def square(a):
    """
    Square the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The squared input.
    """
    return np.multiply(a, a)


def cube(a):
    """
    Cube the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The cubed input.
    """
    return np.multiply(np.multiply(a, a), a)


def sin(a):
    """
    Compute the sine of the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The sine of the input.
    """
    return np.sin(a)


def cos(a):
    """
    Compute the cosine of the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The cosine of the input.
    """
    return np.cos(a)


def tan(a):
    """
    Compute the tangent of the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The tangent of the input.
    """
    return np.tan(a)


def sqrt(a):
    """
    Compute the square root of the input.

    Parameters
    ----------
    a : float
        The input.

    Returns
    -------
    np.ndarray:
        The square root of the input.
    """
    return np.sqrt(np.abs(a))


class SymbolicFunction:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self, func: Callable, arg_count: int, format_str: str, name: str):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = func
        self.arg_count = arg_count
        self.format_str = format_str
        self.name = name

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicAdd:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = np.add
        self.arg_count = 2
        self.format_str = "({} + {})"
        self.name = "add"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicSubtract:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = np.subtract
        self.arg_count = 2
        self.format_str = "({} - {})"
        self.name = "subtract"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicMultiply:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = np.multiply
        self.arg_count = 2
        self.format_str = "({} + {})"
        self.name = "mult"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicDivide:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = safe_div
        self.arg_count = 2
        self.format_str = "({} / {})"
        self.name = "div"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicAbs:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = np.abs
        self.arg_count = 1
        self.format_str = "abs({})"
        self.name = "abs"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicNegate:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = negate
        self.arg_count = 1
        self.format_str = "-({})"
        self.name = "negate"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicSin:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = sin
        self.arg_count = 1
        self.format_str = "sin({})"
        self.name = "sin"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicCos:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = cos
        self.arg_count = 1
        self.format_str = "cos({})"
        self.name = "cos"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicTan:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = tan
        self.arg_count = 1
        self.format_str = "tan({})"
        self.name = "tan"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicSqrt:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = sqrt
        self.arg_count = 1
        self.format_str = "sqrt({})"
        self.name = "sqrt"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicSquare:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = square
        self.arg_count = 1
        self.format_str = "square({})"
        self.name = "square"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


class SymbolicCube:
    """
    An internal class used to represent the symbolic functions used to generate new
    features by the GeneticFeatureGenerator class.
    """

    def __init__(self):
        """
        Initialize the SymbolicFunction class.

        Parameters
        ----------
        func : function
            The function to use.

        arg_count : int
            The number of arguments the function takes.

        format_str : str
            The format string for the function.
        """
        self.func = cube
        self.arg_count = 1
        self.format_str = "cube({})"
        self.name = "cube"

    def __call__(self, *args):
        return self.func(*args)

    def __str__(self):
        return self.format_str


# class SymbolicAddConstant:
#     """
#     An internal class used to represent the symbolic functions used to generate new
#     features by the GeneticFeatureGenerator class.
#     """

#     def __init__(self):
#         """
#         Initialize the SymbolicFunction class.
#         """

#         self.random_constant = np.random.uniform(-1000, 1000)
#         self.arg_count = 1
#         self.format_str = f"add_constant({self.random_constant} + {{}})"
#         self.name = "add_constant"

#     def __call__(self, x):
#         return self.random_constant + x

#     def __str__(self):
#         return self.format_str


# class SymbolicMulConstant:
#     """
#     An internal class used to represent the symbolic functions used to generate new
#     features by the GeneticFeatureGenerator class.
#     """

#     def __init__(self):
#         """
#         Initialize the SymbolicFunction class.
#         """

#         self.random_constant = np.random.uniform(-1000, 1000)
#         self.arg_count = 1
#         self.format_str = f"mul_constant({self.random_constant} + {{}})"
#         self.name = "mul_constant"

#     def __call__(self, x):
#         return self.random_constant * x

#     def __str__(self):
#         return self.format_str


operations = [
    SymbolicAdd,
    SymbolicSubtract,
    SymbolicMultiply,
    SymbolicDivide,
    # SymbolicSqrt,
    SymbolicSquare,
    SymbolicCube,
    SymbolicAbs,
    SymbolicNegate,
    SymbolicSin,
    SymbolicCos,
    SymbolicTan,
]


def list_operations() -> List[str]:
    """
    List the available operations.

    Returns
    -------
    list
        The list of built-in operations.
    """
    return [op().name for op in operations]



""" The population module contains the classes for the population of programs in the
genetic programming algorithm. """

from copy import deepcopy
from typing import Callable, List, Self

import numpy as np
import pandas as pd
from joblib import Parallel, cpu_count, delayed

from .program import random_prog, select_random_node, node_count
from .symbolic_functions import SymbolicFunction


class BasePopulation:
    """
    A class to represent the population of symbolic programs in the
    genetic programming algorithm.
    """

    def __init__(
        self,
        population_size: int,
        operations: List[SymbolicFunction],
        tournament_size: int = 3,
        crossover_prob: float = 0.75,
    ):
        """
        Initialize the population.

        Args
        ----
        population_size : int
            The size of the population. The larger the population, the more
            likely the algorithm will find a good solution, but the longer it
            will take to run.

        operations : list
            The list of functions to use in the programs. These are the
            functions that the algorithm can use to create the programs with.
        """
        self.population_size = population_size
        self.operations = operations
        self.population = None
        self.tournament_size = tournament_size
        self.crossover_prob = crossover_prob

    def initialize(self, X: pd.DataFrame) -> Self:
        """
        Setup the initial population with simple, random programs.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.
        """
        self.population = [
            random_prog(0, X, self.operations) for _ in range(self.population_size)
        ]
        return self

    def evaluate(self, X: pd.DataFrame) -> List[pd.Series]:
        """
        Evaluate the population against the dataframe of features.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.

        return
        ------
        list
            The predicted values.
        """
        raise NotImplementedError

    def compute_fitness(
        self,
        fitness_func: Callable,
        parsimony_coefficient: float,
        prediction,
        y: pd.Series,
    ) -> List[float]:
        """
        Compute the fitness of the population.

        Args
        ----

        fitness_func : callable
            The fitness function to use.

        parsimony_coefficient : float
            The parsimony coefficient.

        prediction : list
            The predicted values.

        y : pd.Series
            The true values.

        return
        ------
        list
            The fitness of the population.
        """
        raise NotImplementedError

    def _evaluate_df(self, node: dict, X: pd.DataFrame) -> pd.Series:
        """
        Evaluate the program against the dataframe of features.

        Args
        ----
        node : dict
            The program to evaluate.

        X : pd.DataFrame
            The dataframe with the features.

        return
        ------
        pd.Series
            The predicted values.
        """
        if "children" not in node:
            return X[node["feature_name"]]
        return pd.Series(
            node["func"](*[self._evaluate_df(c, X) for c in node["children"]])
        )

    def _get_random_parent(self, fitness: List[float]) -> dict:
        """
        Select a random parent from the population using tournament selection.

        Args
        ----
        fitness : list
            The fitness values of the population.

        return
        ------
        dict
            The selected parent program.
        """
        tournament_members = [
            np.random.randint(0, self.population_size - 1)
            for _ in range(self.tournament_size)
        ]
        member_fitness = [(fitness[i], self.population[i]) for i in tournament_members]
        return min(member_fitness, key=lambda x: x[0])[1]

    def _crossover(self, selected1: dict, selected2: dict) -> dict:
        """
        Perform crossover mutation between two selected programs.

        Args
        ----
        selected1 : dict
            The first selected program.

        selected2 : dict
            The second selected program.

        return
        ------
        dict
            The offspring program.
        """
        offspring = deepcopy(selected1)
        xover_point1 = select_random_node(offspring, None, 0)
        xover_point2 = select_random_node(selected2, None, 0)
        child_count = len(xover_point1["children"])
        child_idx = 0 if child_count <= 1 else np.random.randint(0, child_count - 1)
        xover_point1["children"][child_idx] = xover_point2
        return offspring

    def _mutate(self, selected: dict, X: pd.DataFrame) -> dict:
        """
        Mutate the selected program by replacing a random node with a new random program.

        Args
        ----
        selected : dict
            The selected program to mutate.

        X : pd.DataFrame
            The dataframe with the features.
        """
        offspring = deepcopy(selected)
        mutate_point = select_random_node(offspring, None, 0)
        child_count = len(mutate_point["children"])
        child_idx = 0 if child_count <= 1 else np.random.randint(0, child_count - 1)
        mutate_point["children"][child_idx] = random_prog(0, X, self.operations)
        return offspring

    def _get_offspring(self, fitness: List[float], X: pd.DataFrame) -> dict:
        """
        Get the offspring of two parents using crossover mutation.

        Args
        ----
        fitness : list
            The fitness values of the population.

        X : pd.DataFrame
            The dataframe with the features.

        return
        ------
        dict
            The offspring program.
        """
        parent1 = self._get_random_parent(fitness)
        if np.random.uniform() < self.crossover_prob:
            parent2 = self._get_random_parent(fitness)
            return self._crossover(parent1, parent2)

        return self._mutate(parent1, X)

    def evolve(self, fitness: List[float], X: pd.DataFrame) -> Self:
        """
        Evolve the population by creating a new generation of programs.

        Args
        ----
        fitness : list
            The fitness values of the population.

        X : pd.DataFrame
        """
        self.population = [
            self._get_offspring(fitness, X) for _ in range(self.population_size)
        ]
        return self


class SerialPopulation(BasePopulation):
    """
    A class to represent the population of programs in the genetic programming algorithm where
    the programs are evaluated serially.
    """

    def __init__(
        self,
        population_size: int,
        operations: List[SymbolicFunction],
        tournament_size: int = 3,
        crossover_prob: float = 0.75,
    ):
        """
        Initialize the population class.

        Args
        ----

        population_size : int
            The size of the population.

        operations : list
            The list of functions to use in the programs.
        """
        super().__init__(population_size, operations, tournament_size, crossover_prob)

    def evaluate(self, X: pd.DataFrame) -> List[pd.Series]:
        """
        Evaluate the population against the current program.
        """
        return [self._evaluate_df(prog, X) for prog in self.population]

    def compute_fitness(
        self,
        fitness_func: Callable,
        parsimony_coefficient: float,
        prediction,
        y: pd.Series,
    ) -> List[float]:
        """
        Compute the fitness of the population.

        Args
        ----

        fitness_func : callable
            The fitness function to use.

        parsimony_coefficient : float
            The parsimony coefficient.

        prediction : list
            The predicted values.

        y : pd.Series
            The true values.

        return
        ------
        list
            The fitness of the population.
        """
        score = [
            fitness_func(prog, parsimony_coefficient, y, pred)
            for prog, pred in zip(self.population, prediction)
        ]
        return score

    def apply_parsimony(
        self, scores: List[float], parsimony_coefficient: float
    ) -> List[float]:
        """
        Apply the parsimony coefficient to the fitness scores.

        Args
        ----
        scores : list
            The fitness scores.

        parsimony_coefficient : float
            The parsimony coefficient.

        return
        ------
        list
            The updated fitness scores.
        """

        def _parsimony_coefficient(loss, prog):
            penalty = node_count(prog) ** parsimony_coefficient
            loss = loss / penalty
            return -loss

        return [
            _parsimony_coefficient(loss, prog)
            for loss, prog in zip(scores, self.population)
        ]


class ParallelPopulation(BasePopulation):
    """
    A class to represent the population of programs in the genetic programming algorithm where
    the programs are evaluated in parallel via joblib.
    """

    def __init__(
        self,
        population_size: int,
        operations: List[SymbolicFunction],
        tournament_size: int = 3,
        crossover_prob: float = 0.75,
        n_jobs: int = -1,
    ):
        """
        Initialize the population class.

        Args
        ----

        population_size : int
            The size of the population.

        operations : list
            The list of functions to use in the programs.

        n_jobs : int
            The number of jobs to use in the parallel evaluation.
        """
        super().__init__(population_size, operations, tournament_size, crossover_prob)
        n_jobs = cpu_count() if n_jobs == -1 else n_jobs

    def evaluate(self, X: pd.DataFrame) -> List[pd.Series]:
        """
        Evaluate the population against the current program. This is done in parallel.
        """
        return Parallel(n_jobs=cpu_count())(
            delayed(self._evaluate_df)(prog, X) for prog in self.population
        )

    def compute_fitness(
        self,
        fitness_func: Callable,
        parsimony_coefficient: float,
        prediction,
        y: pd.Series,
    ) -> List[float]:
        """
        Compute the fitness of the population.

        Args
        ----

        fitness_func : callable
            The fitness function to use.

        parsimony_coefficient : float
            The parsimony coefficient.

        prediction : list
            The predicted values.

        y : pd.Series
            The true values.

        return
        ------
        list
            The fitness of the population.
        """
        return Parallel(n_jobs=cpu_count())(
            delayed(fitness_func)(prog, parsimony_coefficient, y, pred)
            for prog, pred in zip(self.population, prediction)
        )

    def apply_parsimony(
        self, scores: List[float], parsimony_coefficient: float
    ) -> List[float]:
        """
        Apply the parsimony coefficient to the fitness scores.

        Args
        ----
        scores : list
            The fitness scores.

        parsimony_coefficient : float
            The parsimony coefficient.

        return
        ------
        list
            The updated fitness scores.
        """

        def _parsimony_coefficient(loss, prog):
            penalty = node_count(prog) ** parsimony_coefficient
            loss = loss / penalty
            return -loss

        return Parallel(n_jobs=cpu_count())(
            delayed(_parsimony_coefficient)(score, prog)
            for score, prog in zip(scores, self.population)
        )



"""Class for binary genetic algorithm for feature selection."""

import sys
from typing import Callable, Self, Union

import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import pandas as pd
from joblib import cpu_count
from sklearn.base import BaseEstimator, TransformerMixin
from tqdm import tqdm

from .population import ParallelPopulation, SerialPopulation


class GeneticFeatureSelector(BaseEstimator, TransformerMixin):
    """
    The Genetic Feature Selector class uses genetic programming to select the best
    features to minimise a given objective function. This is done by initially
    building a population of naive random selection of the available features.
    The population is then evolved over a number of generations using genetic operators
    such as mutation and crossover to find the best combination of features to minimise
    the output of the objective function.
    """

    def __init__(
        self,
        objective_function: Callable,
        population_size: int = 50,
        max_generations: int = 100,
        crossover_proba: float = 0.9,
        mutation_proba: float = 0.1,
        early_termination_iters: int = 15,
        n_jobs: int = -1,
        pbar: bool = True,
        verbose: bool = False,
    ) -> None:
        """
        Initialize the genetic algorithm.

        Parameters
        ----------
        objective_function : callable
            The cost function to minimize. Must take X and y as input and return a
            float. Note that the function should return a value to minimize so a
            smaller value is better. If you want to maximize a metric, you should
            multiply the output of your objective_function by -1.

        population_size : int
            The number of individuals in the population.

        max_generations : int
            The maximum number of iterations.

        crossover_proba : float
            The probability of crossover.

        mutation_proba : float
            The probability of mutation.

        early_termination_iters : int
            The number of iterations to wait for early termination.

        n_jobs : int
            The number of parallel jobs to run. If -1, use all available cores else uses the
            minimum of n_jobs and cpu_count.

        verbose : bool
            Whether to print progress.
        """
        self.objective_function = objective_function
        self.population_size = population_size
        self.crossover_proba = crossover_proba
        self.mutation_proba = mutation_proba
        self.max_generations = max_generations

        self.early_termination_iters = early_termination_iters
        self.early_termination_counter = 0

        self.history = []

        self.best_genome = None
        self.best_cost = sys.maxsize

        self.is_fitted_ = False

        if n_jobs == -1:
            self.n_jobs = cpu_count()
        else:
            self.n_jobs = n_jobs

        self.verbose = verbose

        self.pbar = pbar

        self.population = None
        self.num_genes = None
        self.selected_columns = None

    def fit_transform(self, X: pd.DataFrame, y: pd.Series) -> Self:
        """
        Fit the genetic algorithm and return the selected features.

        Parameters
        ----------
        X : DataFrame
            The input features.

        y : Series
            The target variable.
        """
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X: pd.DataFrame, y: pd.Series = None) -> pd.DataFrame:
        """
        Transform the input features to the selected features.

        Parameters
        ----------
        X : DataFrame
            The input features.

        y : Series
            The target variable.

        Returns
        -------
        DataFrame
            The selected features.
        """
        if not self.is_fitted_:
            raise ValueError("Must call fit before transform")

        return X[self.selected_columns]

    def fit(self, X: pd.DataFrame, y: pd.Series) -> Self:
        """
        Determine the  optimal feature selection using a genetic algorithm.

        Args
        ----
        X : DataFrame
            The input features.

        y : Series
            The target variable.

        Returns
        -------
        self
        """
        self.num_genes = X.shape[1]

        if self.n_jobs == 1:
            self.population = SerialPopulation(
                self.population_size,
                self.num_genes,
                self.crossover_proba,
                self.mutation_proba,
            )
        else:
            self.population = ParallelPopulation(
                self.population_size,
                self.num_genes,
                self.crossover_proba,
                self.mutation_proba,
                self.n_jobs,
            )

        if self.pbar:
            pbar = tqdm(
                total=self.max_generations, desc="Optimising feature selection..."
            )

        for current_iter in range(self.max_generations):
            scores = self.population.evaluate(self.objective_function, X, y)

            for genome, score in zip(self.population.population, scores):
                # check for the best genome
                if score < self.best_cost:
                    self.best_cost = score
                    self.best_genome = genome
                    self.early_termination_counter = 0

            # check for early termination
            self.early_termination_counter += 1
            if self.early_termination_counter >= self.early_termination_iters:
                if self.verbose:
                    print(
                        f"Early termination at iter {current_iter}, \
                          best error: {self.best_cost:10.6f}"
                    )
                break

            # store history
            self.history.append(
                {
                    "generation": current_iter,
                    "best_score": self.best_cost,
                    "median_score": np.median(scores),
                }
            )

            # evolve the population
            self.population.evolve(scores)

            if self.pbar:
                pbar.update(1)

        self.is_fitted_ = True

        self.selected_columns = X.columns[self.best_genome == 1]

    def plot_history(self, ax: Union[matplotlib.axes._axes.Axes | None] = None):
        """
        Plot the history of the fitness function.

        return
        ------
        None
        """
        if not self.is_fitted_:
            raise ValueError("Must call fit_transform or transform before plot_history")

        if ax is None:
            _, ax = plt.subplots()

        df = pd.DataFrame(self.history)
        df.plot(x="generation", y=["best_score", "median_score"], ax=ax)
        plt.show()



"""Contains the SymbolicFeatureGenerator class."""

import sys
from typing import List, Union
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
from joblib import cpu_count
from sklearn.base import BaseEstimator, TransformerMixin
from tqdm import tqdm

from .fitness import fitness_pearson
from .mrmr import MaxRelevanceMinRedundancy
from .population import ParallelPopulation, SerialPopulation
from .program import render_prog
from .symbolic_functions import SymbolicFunction, operations
from .preprocess import preprocess_data


class GeneticFeatureSynthesis(BaseEstimator, TransformerMixin):
    """
    The Genetic Feature Synthesis class uses genetic programming to generate new
    features using a technique based on Symbolic Regression. This is done by initially
    building a population of naive random formulas that represent transformations of
    the input features. The population is then evolved over a number of generations
    using genetic operators such as mutation and crossover to find the best programs
    that minimize a given fitness function. The best features are then identified using
    a Maximum Relevance Minimum Redundancy (mRMR) algorithm to find those features
    that are most correlated with the target variable while being least correlated with
    each other.
    """

    def __init__(
        self,
        functions: Union[List[SymbolicFunction] | None] = None,
        num_features: int = 10,
        population_size: int = 100,
        max_generations: int = 25,
        tournament_size: int = 3,
        crossover_proba: float = 0.85,
        parsimony_coefficient: float = 0.001,
        early_termination_iters: int = 15,
        return_all_features: bool = True,
        n_jobs: int = -1,
        pbar: bool = True,
        verbose: bool = False,
    ):
        """
        Initialize the Symbolic Feature Generator.

        Args
        ----
        functions : list
            The list of functions to use in the programs. If `None` then all the
            built-in functions are used.

        num_features : int
            The number of best features to generate. Internally, `3 * num_features`
            programs are generated and the
            best `num_features` are selected via Maximum Relevance Minimum Redundancy
            (mRMR).

        population_size : int
            The number of programs in each generation. The larger the population, the
            more likely it is to find a good solution, but the longer it will take.

        max_generations : int
            The maximum number of generations to run. The larger the number of
            generations, the more likely it is to find a good solution, but the longer
            it will take.

        tournament_size : int
            The size of the tournament for selection. The larger the tournament size,
            the more likely it is to select the best program, but the longer it will
            take.

        crossover_proba : float
            The probability of crossover mutation between selected parents in each
            generation.

        parsimony_coefficient : float
            The parsimony coefficient. Larger values penalize larger programs more and
            encourage smaller programs. This helps prevent bloat where the programs
            become increasingly large and complex without improving the fitness, which
            increases computation complexity and reduces the interpretability of the
            features.

        early_termination_iters : int
            If the best score does not improve for this number of generations, then the
            algorithm will terminate early.

        return_all_features : bool
            Whether to return all the features generated or just the best features.

        n_jobs : int
            The number of parallel jobs to run. If `-1`, use all available cores else
            uses n_jobs. If `n_jobs=1`, then the computation is done in serial.

        pbar: bool
            Whether to show a progress bar.

        verbose : bool
            Whether to print out aditional information
        """
        if functions is None:
            self.functions = operations
        else:
            self.functions = functions

        self.population_size = population_size
        self.max_generations = max_generations
        self.tournament_size = tournament_size
        self.crossover_proba = crossover_proba
        self.num_features = num_features
        self.parsimony_coefficient = parsimony_coefficient

        self.history = []
        self.hall_of_fame = []
        self.len_hall_of_fame = self.num_features * 5

        self.population = None

        self.early_termination_iters = early_termination_iters
        self.early_termination_counter = 0

        self.return_all_features = return_all_features

        self.fitness_func = fitness_pearson

        self.verbose = verbose

        self.fit_called = False

        if n_jobs == -1:
            self.n_jobs = cpu_count()
        else:
            self.n_jobs = n_jobs

        self.pbar = pbar

    def _update_hall_of_fame(self, fitness: List[float]):
        for individual, fit in zip(self.population.population, fitness):
            current_fitnesses = [x["fitness"] for x in self.hall_of_fame]
            if fit not in current_fitnesses:
                self.hall_of_fame.append({"individual": individual, "fitness": fit})

        self.hall_of_fame = sorted(self.hall_of_fame, key=lambda x: x["fitness"])
        self.hall_of_fame = self.hall_of_fame[: self.len_hall_of_fame]

    def _select_best_features(self, X: pd.DataFrame, y: pd.Series):
        """
        Select the best features using the mRMR algorithm.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        return
        ------
        None
        """
        population = SerialPopulation(
            len(self.hall_of_fame),
            self.functions,
            self.tournament_size,
            self.crossover_proba,
        )

        population.population = [x["individual"] for x in self.hall_of_fame]
        features = pd.DataFrame(population.evaluate(X)).T

        features.columns = [f"feature_{i}" for i in range(self.len_hall_of_fame)]

        for i in range(self.len_hall_of_fame):
            self.hall_of_fame[i]["name"] = f"feature_{i}"

        selected = (
            MaxRelevanceMinRedundancy(k=self.num_features, pbar=self.pbar)
            .fit_transform(features, y)
            .columns
        )
        selected = [int(x.split("_")[1]) for x in selected]
        self.hall_of_fame = [self.hall_of_fame[i] for i in selected]

    def fit(self, X: pd.DataFrame, y: pd.Series) -> "GeneticFeatureSynthesis":
        """
        Fit the symbolic feature generator to the data.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        return
        ------
        returns self
        """
        X_copy, y_copy = preprocess_data(
            X.reset_index(drop=True), y.reset_index(drop=True)
        )

        # Initialize the population
        if self.n_jobs == 1:
            self.population = SerialPopulation(
                self.population_size,
                self.functions,
                self.tournament_size,
                self.crossover_proba,
            ).initialize(X_copy)
        else:
            self.population = ParallelPopulation(
                self.population_size,
                self.functions,
                self.tournament_size,
                self.crossover_proba,
                self.n_jobs,
            ).initialize(X_copy)

        # loss value to minimize
        global_best = sys.maxsize
        best_prog = None

        if self.pbar:
            pbar = tqdm(total=self.max_generations, desc="Creating new features...")

        for gen in range(self.max_generations):
            fitness = []
            prediction = self.population.evaluate(X_copy)
            score = self.population.compute_fitness(
                self.fitness_func, self.parsimony_coefficient, prediction, y_copy
            )
            # import pdb

            # pdb.set_trace()
            # score = self.population.apply_parsimony(score, self.parsimony_coefficient)
            # prog_len = [node_count(prog) for prog in self.population.population]
            # clf = np.cov(prog_len, score)[0, 1]
            # vl = np.var(prog_len)
            # parsimony = clf / vl
            # score = [x - (parsimony * y) for x, y in zip(score, prog_len)]

            for prog, score in zip(self.population.population, score):

                fitness.append(score)
                if score < global_best:
                    global_best = score
                    best_prog = prog
                    self.early_termination_counter = 0

            # update the history
            results = {
                "generation": gen,
                "best_score": global_best,
                "median_score": pd.Series(fitness).median(),
                "best_program": render_prog(best_prog),
            }
            self.history.append(results)

            # check for early termination
            self.early_termination_counter += 1
            if self.early_termination_counter >= self.early_termination_iters:
                if self.verbose:
                    print(
                        f"Early termination at iter {gen}, best error: {global_best:10.6f}"
                    )
                break

            if self.pbar:
                pbar.update(1)

            # update the hall of fame with the best programs from the current generation
            self._update_hall_of_fame(fitness)

            self.population.evolve(fitness, X_copy)

        # select the best features using mrmr
        self._select_best_features(X_copy, y_copy)

        if self.verbose:
            print("Symbolic Feature Generator")
            print(f"Best program: {render_prog(best_prog)}")
            print(f"Best score: {global_best}")

        # we've successfully finished the fit
        self.fit_called = True

        return self

    def transform(self, X: pd.DataFrame, y: pd.Series = None) -> pd.DataFrame:
        """
        Transform the dataframe of features using the best programs found.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.

        return
        ------
        pd.DataFrame
            The transformed dataframe.
        """
        if not self.fit_called:
            raise ValueError("Must call fit before transform")

        if self.n_jobs == 1:
            population = SerialPopulation(
                len(self.hall_of_fame),
                self.functions,
                self.tournament_size,
                self.crossover_proba,
            )
        else:
            population = ParallelPopulation(
                len(self.hall_of_fame),
                self.functions,
                self.tournament_size,
                self.crossover_proba,
                self.n_jobs,
            )

        population.population = [x["individual"] for x in self.hall_of_fame]
        output = pd.DataFrame(population.evaluate(X.reset_index(drop=True))).T
        output.columns = [x["name"] for x in self.hall_of_fame]

        if self.return_all_features:
            return pd.concat([X.reset_index(drop=True), output], axis=1)

        return output

    def fit_transform(self, X: pd.DataFrame, y: pd.Series = None) -> pd.DataFrame:
        """
        Fit the symbolic feature generator to the data and transform the dataframe of features.

        Args
        ----
        X : pd.DataFrame
            The dataframe with the features.

        y : pd.Series
            The target variable.

        return
        ------
        pd.DataFrame
            The transformed dataframe.
        """
        self.fit(X, y)
        return self.transform(X, y)

    def get_feature_info(self) -> pd.DataFrame:
        """
        Get the information about the best programs found.

        return
        ------
        pd.DataFrame
            The dataframe with the information.
        """
        if not self.fit_called:
            raise ValueError("Must call fit before get_feature_info")

        output = []
        for prog in self.hall_of_fame:
            tmp = {
                "name": prog["name"],
                "formula": render_prog(prog["individual"]),
                "fitness": prog["fitness"],
            }
            output.append(tmp)

        return pd.DataFrame(output)

    def plot_history(self, ax: Union[matplotlib.axes._axes.Axes | None] = None):
        """
        Plot the history of the fitness function.

        return
        ------
        None
        """
        if not self.fit_called:
            raise ValueError("Must call fit before plot_history")

        if ax is None:
            _, ax = plt.subplots()

        df = pd.DataFrame(self.history)
        df.plot(x="generation", y=["best_score", "median_score"], ax=ax)
        plt.show()



import featuristic as ft
import numpy as np
import pandas as pd
from pytest import approx


def test_symbolic_addition():
    func = ft.synthesis.symbolic_functions.SymbolicAdd()
    assert func(1, 2) == 3

    a = pd.Series([1, 2])
    b = pd.Series([3, 4])
    c = pd.Series([4, 6])
    assert (func(a, b) == c).all()

    a = np.array([1, 2])
    b = np.array([3, 4])
    c = np.array([4, 6])
    assert (func(a, b) == c).all()


def test_symbolic_subtraction():
    func = ft.synthesis.symbolic_functions.SymbolicSubtract()
    assert func(3, 2) == 1

    a = pd.Series([1, 2])
    b = pd.Series([3, 5])
    c = pd.Series([2, 3])
    assert (func(b, a) == c).all()

    a = np.array([1, 2])
    b = np.array([3, 5])
    c = np.array([2, 3])
    assert (func(b, a) == c).all()


def test_symbolic_multiply():
    func = ft.synthesis.symbolic_functions.SymbolicMultiply()
    assert func(3, 2) == 6

    a = pd.Series([2, 3])
    b = pd.Series([3, 5])
    c = pd.Series([6, 15])
    assert (func(b, a) == c).all()

    a = np.array([2, 3])
    b = np.array([3, 5])
    c = np.array([6, 15])
    assert (func(b, a) == c).all()


def test_symbolic_divide():
    func = ft.synthesis.symbolic_functions.SymbolicDivide()
    assert func(4, 2) == 2

    a = pd.Series([12, 10])
    b = pd.Series([3, 5])
    c = pd.Series([4, 2])
    assert (func(a, b) == c).all()

    a = np.array([12, 10])
    b = np.array([3, 5])
    c = np.array([4, 2])
    assert (func(a, b) == c).all()


def test_symbolic_abs():
    func = ft.synthesis.symbolic_functions.SymbolicAbs()
    assert func(-1) == 1

    a = pd.Series([1, -2])
    b = pd.Series([1, 2])
    assert (func(a) == b).all()

    a = np.array([1, -2])
    b = np.array([1, 2])
    assert (func(a) == b).all()


def test_symbolic_negate():
    func = ft.synthesis.symbolic_functions.SymbolicNegate()
    assert func(1) == -1

    a = pd.Series([1, -2])
    b = pd.Series([-1, 2])
    assert (func(a) == b).all()

    a = np.array([1, -2])
    b = np.array([-1, 2])
    assert (func(a) == b).all()


def test_symbolic_square():
    func = ft.synthesis.symbolic_functions.SymbolicSquare()
    assert func(2) == 4

    a = pd.Series([2, 4, 8])
    b = pd.Series([4, 16, 64])
    assert (func(a) == b).all()

    a = np.array([2, 4, 8])
    b = np.array([4, 16, 64])
    assert (func(a) == b).all()


def test_symbolic_cube():
    func = ft.synthesis.symbolic_functions.SymbolicCube()
    assert func(2) == 8

    a = pd.Series([2, 4, 8])
    b = pd.Series([8, 64, 512])
    assert (func(a) == b).all()

    a = np.array([2, 4, 8])
    b = np.array([8, 64, 512])
    assert (func(a) == b).all()


def test_symbolic_sin():
    func = ft.synthesis.symbolic_functions.SymbolicSin()
    assert func(1) == approx(0.84147098)

    a = pd.Series([0, 0.25, 0.5, 0.75, 1])
    b = pd.Series(
        [
            approx(0.0),
            approx(0.24740396),
            approx(0.47942554),
            approx(0.68163876),
            approx(0.84147098),
        ]
    )
    assert (func(a) == b).all()

    a = np.array([0, 0.25, 0.5, 0.75, 1])
    b = np.array(
        [
            approx(0.0),
            approx(0.24740396),
            approx(0.47942554),
            approx(0.68163876),
            approx(0.84147098),
        ]
    )
    assert (func(a) == b).all()


def test_symbolic_cos():
    func = ft.synthesis.symbolic_functions.SymbolicCos()
    assert func(1) == approx(0.54030231)

    a = pd.Series([0, 0.25, 0.5, 0.75, 1])
    b = pd.Series(
        [
            approx(1.0),
            approx(0.96891242),
            approx(0.87758256),
            approx(0.73168887),
            approx(0.54030231),
        ]
    )
    assert (func(a) == b).all()

    a = np.array([0, 0.25, 0.5, 0.75, 1])
    b = np.array(
        [
            approx(1.0),
            approx(0.96891242),
            approx(0.87758256),
            approx(0.73168887),
            approx(0.54030231),
        ]
    )
    assert (func(a) == b).all()


def test_symbolic_tan():
    func = ft.synthesis.symbolic_functions.SymbolicTan()
    assert func(1) == approx(1.55740772)

    a = pd.Series([0, 0.25, 0.5, 0.75, 1])
    b = pd.Series(
        [
            approx(0.0),
            approx(0.25534192),
            approx(0.54630249),
            approx(0.93159646),
            approx(1.55740772),
        ]
    )
    assert (func(a) == b).all()

    a = np.array([0, 0.25, 0.5, 0.75, 1])
    b = np.array(
        [
            approx(0.0),
            approx(0.25534192),
            approx(0.54630249),
            approx(0.93159646),
            approx(1.55740772),
        ]
    )
    assert (func(a) == b).all()


def test_symbolic_sqrt():
    func = ft.synthesis.symbolic_functions.SymbolicSqrt()
    assert func(1) == approx(1.0)

    a = pd.Series([0, 0.25, 0.5, 0.75, 1])
    b = pd.Series(
        [
            approx(0.0),
            approx(0.5),
            approx(0.707107),
            approx(0.866025),
            approx(1.0),
        ]
    )

    assert (func(a) == b).all()

    a = np.array([0, 0.25, 0.5, 0.75, 1])
    b = np.array(
        [
            approx(0.0),
            approx(0.5),
            approx(0.707107),
            approx(0.866025),
            approx(1.0),
        ]
    )
    assert (func(a) == b).all()


def test_list_operations():
    ops = ft.synthesis.symbolic_functions.list_operations()

    assert len(ops) > 0
    assert all(isinstance(op, str) for op in ops)
